/*
        fr1.c
        Filter to copy a RGB file to a scsi port with a Montage
		FR-1 slide printer connected to it
		
        Written by Mark A. Haidekker, mark@haidekker.org
        Released under the terms of the GPL.
		
		
		DISCLAIMER
		
This software is (c) 2000-2001 Mark A. Haidekker.

You may use this software under the terms and conditions of the
GPL - see attached file "COPYING"

By using this software you agree to be bound by the therms of the GPL
and, in eccess of GPL, by:

1. DISCLAIMER OF WARRANTY.  THE PRODUCT IS PROVIDED FREE OF CHARGE,
AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF ANY KIND,
INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT IS FREE OF
DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE OR
NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE
OF THE PRODUCT IS BORNE BY THE USER.  SHOULD THE PRODUCT PROVE
DEFECTIVE IN ANY RESPECT, THE USER AND NOT THE AUTHOR
ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  NO
USE OF THE PRODUCT IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS
DISCLAIMER.

2. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW, IN NO EVENT WILL LICENSOR OR ITS SUPPLIERS OR
RESELLERS BE LIABLE FOR ANY INDIRECT, SPECIAL, INCIDENTAL OR
CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF OR INABILITY TO USE
THE PRODUCT, INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF
GOODWILL, WORK STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY
AND ALL OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF, AND REGARDLESS OF THE LEGAL OR EQUITABLE
THEORY (CONTRACT, TORT OR OTHERWISE) UPON WHICH THE CLAIM IS BASED.
THE AUTHOR IS NOT RESPONSIBLE FOR
ANY LIABILITY ARISING OUT OF CONTENT PROVIDED BY THE USER OR A
THIRD PARTY THAT IS ACCESSED THROUGH THE PRODUCT AND/OR ANY
MATERIAL LINKED THROUGH SUCH CONTENT.

If you do not agree to the above, delete this software immediately.

		
	
        THIS IS ALPHA SOFTWARE.  IT WORKS FINE FOR ME.
        IT MIGHT NOT WORK FOR YOU.  You'll need to be
        familiar with rebuilding kernels to take
        advantage of this program (under older versions
		of Linux at least). 
		
		Some code from hpscanpbm by
		David Etherton, etherton@netcom.com has been used,
		most importantly (but not exclusively) sg_xfer.
		Thanks, David. It helped a lot!

        Build with cc -o fr1 fr1.c

        THIS IS NOT A DEVICE DRIVER. Your FR-1 must be connected
        to a SCSI adapter that is supported by Linux. I am using the
		Adaptec 3920 UW with a dual SCSI bus having internal drives
		on bus B and external devices (FR-1) on bus A.

        This program uses the generic SCSI interface.  Therefore, you 
        must have generic scsi enabled in your kernel, and you should
        have a series of generic SCSI devices set up:
        
                /dev/sga        major 21, minor 0 (mknod /dev/sga c 21 0)
                /dev/sgb        major 21, minor 1 (mknod /dev/sgb c 21 1)
                :               :

        SCSI generic devices should be character devices, not block
        devices.  Furthermore, even devices that are already assigned
        to other names (like /dev/sda, /dev/sr0) also get a generic
        entry point.  This means that you need at least as many scsi
        generic device entries as you have scsi devices.
		
		The Montage FR-1 does not respond correctly to the inquiry command.
		Thus, the FR-1 must be turned on AND FULLY WARMED UP at boot-
		time of the linux box, otherwise, the generic device entry
		will not be created.
		
		The generic SCSI device (e.g. /dev/sga) must have rw-rw-rw
		permissions.

*/


#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <scsi/sg.h>		/* /usr/src/linux/include/scsi; /usr/include/scsi */



#define LUN			0x00
							/* 0x60 is LUN 3 shifted to the 3 MSBs */

static struct sg_header hdr; 


char version[80];	/* Version string */
char devname[255];	/* Device, e.g. /dev/sga */
char fname[255];	/* Input file name */
char tmpname[255];	/* Name of temp directory */
FILE* inp;			/* binary ppm input file */
FILE* thnail;		/* Binary ppm thumbnail */
long maxlen;
int resolution;		/* 0 for low (2048 x 1365), 1 for high (4096 by 2731) */
int xmax, ymax;		/* Set according to the above */
int lowcontrast;	/* set to 1 if low contrast wanted */
int finecolor;		/* set to 1 if fine color shading (and slower) */
int filmtype;		/* the film type 0 - 6 */
int mirror;			/* If image is to be mirrored */
int forceoper;		/* Force continuation of operation through fatal errors */
char debugmsg = 1;		/* Debug messages */
char debugdump=1;	/* Dump buffer contents */

int rows, cols, maxgray;	/* info we get from the ppm header */
char flip;					/* if rows (x) < cols (y), flip (portrait) */
unsigned char* PANE;		/* one color pane array */
int slideno;				/* Slide number being exposed */
long hdrend;				/* File position at end of header */
double scalefac;			/* interpolation scaling factor */
double padx, pady;			/* To pad borders if non-1.5 aspect ratio */
char autoframe;				/* Flag whether a white border should be detected */
int begx,begy;				/* Start of image if white border detected */
int endx, endy;				/* End of image when white border detected */
int deltax, deltay;			/* additional spacers for auto-frame mode */
int thumb_x, thumb_y;		/* Thumbnail resolution */

char memhog;		/* If 1, buffer all 3 panes (faster) */
char testmode;		/* Dump the panes as pgm in /var/tmp for review */
char traceflg;		/* If ON, create thumbnails for all exposed slides */


/************************************************************************/

int min(int a,int b) { return a<b? a : b; }

double fmin(double a,double b) { return a<b? a : b; }
double fmax(double a,double b) { return a>b? a : b; }


int norm8(double value)
{
	if (value < -1.0)
		value = -1.0;
	else if (value > 1.0)
		value = 1.0;
	return (int)(value * 127.0);
}

char isempty (char* s)
{
int i;

	i=0;
	while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t') || (s[i]=='\n'))) i++;
	if (i<strlen(s)) return 0; else return 1;
}



/* This is for debugging: Print stuff to stderr */

int dp (char *format,...)
{
va_list args;
char buf[1024];

	if (!debugmsg) return 0;
    va_start(args, format);
    vsprintf(buf,format,args);
    fprintf (stderr,buf,strlen(buf));
	return 0;
}


/* This is also for debugging: Dump count bytes in buffer to console */

static void dump(void *buffer,int count)
{
unsigned char *c = buffer;
int i,j;
    
	if (!debugdump) return;
	
	for (i=0; i<count; i+=16) 
	{
    	printf("%04x: ",i);
        for (j=i; j<i+16; j++)
			if (j < count)
           		printf("%02x ",c[j]);
            else
                printf("   ");
            for (j=i; j<i+16; j++)
                if (j < count)
                	putchar(isprint(c[j])?c[j]:'.');
                else
                	putchar(' ');
                
				putchar('\n');
        }
}


/* The following handles the SCSI requests */

    
int sg_xfer(int fid,unsigned char *src,int len,unsigned char *dest,
                int reply_len)
{
int count, result;
/* static struct sg_header hdr; we define this globally to allow universal access */
static unsigned char buf[5012 + sizeof(struct sg_header)];
long i;

	for (i=0; i<100000; i++) ;

	hdr.pack_len = sizeof(hdr) + len;
    hdr.reply_len = sizeof(hdr) + reply_len;
    memcpy(buf, &hdr, sizeof(hdr));
    memcpy(buf + sizeof(hdr), src, len);
    if (write(fid, buf, hdr.pack_len) == -1)
	{
    	perror("sg_xfer(write)");
        return -1;
    }

    if ((count = read(fid,buf,hdr.reply_len)) == -1) 
	{
     	perror("sg_xfer(read)");
        return -1;
    }

    result = ((struct sg_header*)buf)->result;
    
	if (result) 
	{
    	fprintf(stderr,"sg_xfer: result is %d\n",result);
    }
        
	if (count > sizeof(struct sg_header))
	{
      	count -= sizeof(struct sg_header);
     	memcpy(dest,buf + sizeof(hdr),count);
     }
     else
     	count = 0;
     return count;
}



/* 

Write len bytes of data to the device handle fid
This handles the first 5 bytes for the fr_command.
Upon return, look into hdr->target_status and hdr->sense_buffer

*/


int obsolete_write_data(int fid, char *data, int len)  /* obsolete */
{
char write_6[5012];  /* must allow for 4096 bytes data plus header */

   	memset(write_6,0,5);
    write_6[0] = 0xa;
	write_6[1] = LUN;
  	write_6[2] = len >> 16;
  	write_6[3] = len >> 8;
  	write_6[4] = len;
 	memcpy(write_6 + 5, data, len); /* append data to write_6 */

	if (sg_xfer(fid,write_6,5 + len,0,0) < 0) return -1;
    return 0;
}

/* Read a maximum of len bytes from device fid, return # of bytes read */

int obsolete_read_data(int fid,char *dest,int len)   /* obsolete */
{
static unsigned char read_6[6] = { 8,LUN,0,0,0,0 };
    read_6[2] = len >> 16;
    read_6[3] = len >> 8;
    read_6[4] = len;
    if ((len=sg_xfer(fid,read_6,6,dest,len)) < 0) return -1;
    return len;
}


/* This writes a number of strings (thus ...) to fid */

int obsolete_write_string(int fid,char *format,...) /* obsolete */
{
va_list args;
char buf[1024];

    va_start(args, format);
    vsprintf(buf,format,args);
    obsolete_write_data(fid,buf,strlen(buf));
	return 0;
}



/* 

Now we are getting to the real stuff. The FR-1 PRINT command 
transfers parameters to the FR-1 such as film type or raster data.
PRINT request buffer starts with a 0x0A, followed by 0x00.
The next three bytes are the transfer length (MSB - LSB)
followed by either 0x00 or 0x40.

0x00 indicates that a FR command follows
0x40 indicates that source data follows.

A test_unit_ready or inquire_status should be performed before using this

*/


void fr_command (int fid, int code, long parm)
{
unsigned char buf[256];
long i,j;

	j=0;
	/* for (i=0; i<6553600; i++) j=j+i; */
	for (i=0; i<256; i++) buf[i]=0;
	buf[0] = 0x0a;		/* Print command, total of 6 bytes for the SCSI core */
	buf[1] = LUN;
	buf[2] = 0;			/* transfer length MSB */
	buf[3] = 0;			/* transfer length  */
	buf[4] = 5;			/* transfer length LSB */
	buf[5] = 0x00;		/* a command follows, 5 additional bytes */
	
	buf[ 6] = code & 0xff;
	buf[ 7] = (parm >> 24) & 0xff;
	buf[ 8] = (parm >> 16) & 0xff;
	buf[ 9] = (parm >>  8) & 0xff;
	buf[10] = (parm      ) & 0xff;

	if (sg_xfer(fid, buf, 11 ,0,0) < 0) dp ("Error writing FR command\n");
}





/*

Send a SCSI Inquiry command. The request field starts with 0x12,
followed by 3 zeros, then the reply buffer length, then the 
inquiry type (most significant 2 bits) as follows:

00 (0x00)	Returns value (depending on command 31) such as errors
01 (0x40)	Returns model & firmware ID string
10 (0x80)	Status
11 (0xC0)	Never tried it. Probably illegal

The three handlers for 0x00, 0x40, 0x80 are get_ready_status, inquire_model,
and inquire_status.

Model and firmware ID are (function 0x40):
	Byte 5 - 24		Identification string
	Byte 25 - 29	Version number, format xx.yy
	Byte 31 - 40	Date of firmware release
	Bytes 42 - 		Copyright message or other superfluous junk	
	
Status bytes are (function 0x80):
	Byte 5:		MSB contains flag (1 if raster data may be sent)
	Byte 6,7,8:	Maximum source raster transfer length MSB-LSB, 24 bits

Error/ready bytes are (function 0x00):
	0 if FR not ready
	1 if FR ready to make an image
	2 if the FR is out of film
	3 if the FR has no camera back attached


*/


int inquire_model(int fid, char *id, char *version)
{
static unsigned char inquiry[] = { 0x12,LUN,0,0,255, 0x40 };
unsigned char reply[256];
int rtnlen;	// Bytes returned
       
	memset (reply,0xff,sizeof(reply));
	rtnlen= sg_xfer(fid,inquiry,sizeof(inquiry),reply,sizeof(reply));
	dp ("Inquire_status: sg_xfer returned %d bytes\n", rtnlen);
	if (reply[0]!=2) dp ("Warning: Inquiry possibly failed. Returned value %d\n",reply[0]);
    memcpy(id, reply + 5, 20);
    memcpy(version,reply + 25, 5);
    return 0;
}

int inquire_status (int fid)
{
static unsigned char inquiry[] = { 0x12,LUN,0,0,255, 0x80 };
unsigned char reply[256];
int rtnlen;	// Bytes returned
       
	memset (reply,0xff,sizeof(reply));
	rtnlen= sg_xfer(fid,inquiry,sizeof(inquiry),reply,sizeof(reply));
	/*dp ("Inquire_status: sg_xfer returned %d bytes\n", rtnlen);*/
	if (reply[0]!=2) dp ("Warning: Inquiry possibly failed. Returned value %d\n",reply[0]);

	maxlen = (((reply[6] <<8) + reply[7]) <<8 ) + reply[8];
	/*dp ("Inquire_status: %d bytes may be sent\n", maxlen);*/
	if (reply[5] & 0x80)
		return 1;		 /* source raster information may be sent */
	else
		return 0;		/* fatal error or not between begin/end */
}

int get_ready_status (int fid)
{
static unsigned char inquiry[] = { 0x12,LUN,0,0,255, 0x00 };
unsigned char reply[256];
int rtnlen;	// Bytes returned
int i, rtnparm;       

	memset (reply,0xff,sizeof(reply));
	fr_command (fid, 31, 0); /* FR value #0 returns the ready status  */
	rtnlen= sg_xfer(fid,inquiry,sizeof(inquiry),reply,sizeof(reply));
	if (reply[0]!=2) dp ("Warning: Inquiry possibly failed. Returned value %d\n",reply[0]);
	rtnparm=0;
	for (i=0; i<reply[4]; i++) rtnparm = (rtnparm<<8) + reply[i+5];
    return rtnparm;
}


/*

Send len bytes of raw data to the FR
Do an inquire_status first to see if the FR-1 can accept the data

*/

void fr_data (int fid, unsigned char* buf, int len)
{
unsigned char buf1 [4200];   /* must accomodate buf + header */
int i,j;


	do
	{
		i = inquire_status (fid);
		if (i==0)
		{
			for (j=0; j<4096; j++);
			/* dp ("Waiting for FR-1 to accept data. %d of %d would be accepted\n",maxlen,len); */
		}
	}
	while ((i==0) && (len > maxlen));

	buf1[0] = 0x0a;		/* Print command, total of 6 bytes for the SCSI core */
	buf1[1] = LUN;
	buf1[2] = (len >> 16) & 0xff;			/* transfer length MSB */
	buf1[3] = (len >>  8) & 0xff;			/* transfer length  */
	buf1[4] = (len      ) & 0xff;			/* transfer length LSB */
	buf1[5] = 0x40;		/* raster data follows, len additional bytes */

 	memcpy(buf1+6, buf, len); /* append data to buf1 */
	if (sg_xfer(fid, buf1, 6+len ,0,0) < 0) dp ("Error writing FR command\n");
}





/*

TEST_UNIT_READY: This command is used to determine if the FR-1 can accept
the next SCSI Print command. The request buffer starts with 0x03
(for test_unit_ready) followed by 0 0 0 0, then either 0x00 or 0x40.

Using 0x40, this tests for readyness.

If inquire_status returned 00 (cant accept data), use 0x00 as
the last byte. This indicates a fatal error or a necessary wait.

*/

int test_unit_ready (int fid)
{
static unsigned char t_u_r[] = { 0x00,LUN,0,0,0, 0 };
unsigned char reply[256];
int rtnlen;	// Bytes returned
int i;
      
	  
	for (i=0; i<256; i++) reply[i]=0;   
	t_u_r[5] = 0x40;
	rtnlen= sg_xfer(fid,t_u_r,sizeof(t_u_r),0,0);
	
	dp ("Target_status is %d\n", hdr.target_status);
		
	return hdr.target_status;
}


/* Use the above to wait until readyness is signalled */

void wait_for_ready (int fid)
{
static unsigned char rotor[4]={'-','\\','|','/'};
int rotcnt;
long dly;

	rotcnt = 0;
	// dp ("Waiting for unit to be ready...\n");
	while (test_unit_ready (fid))
	{
		dp ("%c\b",rotor[(rotcnt++) &0x03]);
		for (dly=0; dly<4096; dly++);		/* crude but works somehow */
	}
	dp (" ");
	/* dp ("Unit is now ready.\n"); */
}


/*

REQUEST SENSE: Quick-check if any errors occurred
Request-sense returns a 4-byte header where bits 6-5-4
contain the error class:
0	No error
1	Non fatal warnings
2	Fatal errors

*/

int request_sense (int fid)
{
static unsigned char rq_s[] = { 0x03,LUN,0,0,4, 0 };
unsigned char reply[256];
int rtnlen;	// Bytes returned
int i;
      
	for (i=0; i<256; i++) reply[i]=0;
	rtnlen= sg_xfer(fid,rq_s,sizeof(rq_s),reply,sizeof(reply));
	i = reply[0] >> 4;
	if (i) dp ("Request_sense: Error class returned is %d\n",i);
	return i;
}




/*

As above, we use the "get FR value" command to read the error flags
and also read out the FIFO of error codes

*/

int get_errors (int fid)
{
static unsigned char inquiry[] = { 0x12,LUN,0,0,255, 0 };
unsigned char reply[256];
int rtnlen;	// Bytes returned
int i, rtnparm, cnt, rtncode;

	rtncode=0;
	fr_command (fid, 31, 1); /* FR value #1 returns the fatal errors */
	rtnlen= sg_xfer(fid,inquiry,sizeof(inquiry),reply,sizeof(reply));
	if (reply[0]!=2) dp ("Warning: Inquiry possibly failed. Returned value %d\n",reply[0]);
	rtnparm=0;
	for (i=0; i<reply[4]; i++) rtnparm = (rtnparm<<8) + reply[i+5];
	dp ("Get_errors: Number of fatal errors: %d\n", rtnparm);
	if (rtnparm>0) rtncode|=2;

	fr_command (fid, 31, 2); /* FR value #2 returns the warnings */
	rtnlen= sg_xfer(fid,inquiry,sizeof(inquiry),reply,sizeof(reply));
	if (reply[0]!=2) dp ("Warning: Inquiry possibly failed. Returned value %d\n",reply[0]);
	rtnparm=0;
	for (i=0; i<reply[4]; i++) rtnparm = (rtnparm<<8) + reply[i+5];
	dp ("Get_errors: Number of warnings: %d\n", rtnparm);
	if (rtnparm>0) rtncode|=1;
	
	cnt=0;
	dp ("Error code FIFO: ");
	do
	{
		fr_command (fid, 31, 4); /* FR value #4 pops & returns the error FIFO */
		rtnlen= sg_xfer(fid,inquiry,sizeof(inquiry),reply,sizeof(reply));
		if (reply[0]!=2) dp ("Warning: Inquiry possibly failed. Returned value %d\n",reply[0]);
		rtnparm=0;
		for (i=0; i<reply[4]; i++) rtnparm = (rtnparm<<8) + reply[i+5];
		dp ("%d ", rtnparm);
	}
	while ((rtnparm) && (cnt++<30));
	dp ("\n");
	return rtncode;

}



/*====================================================================*/

void testimage (int fid)
{
int x,y;
unsigned char buf[4096];
unsigned char ftype;

	/* start with a reset (initialize, command 0).
		errors should be cleared. If errors occur, we may safely
		exit the program because it won't work anyway. Errors
		of this type include camera back not attached, or any of the
		pins 4 and 7 pulled low
	*/

	fr_command (fid, 0,0);
	switch (request_sense(fid))
	{
		case 0: dp ("Request_sense (1) indicated no errors\n"); break;
		case 1: dp ("Request_sense (1) indicated warnings\n"); 
				get_errors(fid);
				break;
		case 2: dp ("Request_sense (1) indicated fatal errors\n");
				get_errors(fid);
				if (!forceoper) return;
				break;
	}


	
	/* Set up parameters */
	
	fr_command (fid, 3, resolution);
	
	/* Film type: 	0 = default Ektachrome 100
					1 = Ektachrome 100
					2 = Polachrome instant 35 mm
					3 = Polaroid 339 paper film
					4 = Polaroid 669
					5 = Polaroid 691
					6 = Kodacolor 100
					
		OR into this a 16 if fine color selected
		OR into this a 32 if low contrast selected
	*/
	
	ftype = filmtype;
	if (finecolor) ftype |= 16;
	if (lowcontrast) ftype |= 32;
	fr_command (fid, 4, ftype);
	
	fr_command (fid, 7, 1);	/* Test: Set mirror - works! */

	/* Quick-check for errs, 1=warning, 2=fatal 
	   Abort operation if any fatals occured during set-up */

	switch (request_sense(fid))
	{
		case 0: dp ("Request_sense (2) indicated no errors\n"); break;
		case 1: dp ("Request_sense (2) indicated warnings\n"); 
				get_errors (fid);
				break;
		case 2: dp ("Request_sense (2) indicated fatal errors\n");
				get_errors (fid);
				if (!forceoper) return;
				break;
	}
	
	switch (get_ready_status (fid))
	{
		case 0: dp ("Get_ready_status: Camera not ready for unknown reason\n");
				/* if (!forceoper) return; */
				break;
		case 1:	dp ("Get_ready_status: Device is ready.\n");
				break;
		case 2: dp ("Get_ready_status: Camera not ready - out of film\n");
				if (!forceoper) return;
				break;
		case 3: dp ("Get_ready_status: Camera not ready - No camera back attached\n");
				if (!forceoper) return;
				break;
	}
				
	
	

	/* Do the actual exposure */
	
	fr_command (fid, 5,0);   /* start exposure */
	
	fr_command (fid, 8,0);	/* begin color red */
	dp ("Exposing red...");
	for (y=0; y<ymax; y++)
	{
		for (x=0; x<xmax; x++) buf[x]=(x) & 0xff;
		fr_data (fid, buf, xmax);
	}
	fr_command (fid, 9,0);	/* end color */
	dp ("done\n");

	fr_command (fid, 8,1);	/* begin color green */
	dp ("Exposing green...");
	for (y=0; y<ymax; y++)
	{
		for (x=0; x<xmax; x++) buf[x]=(y) & 0xff;
		fr_data (fid, buf, xmax);
	}
	fr_command (fid, 9,0);	/* end color */
	dp ("done\n");

	fr_command (fid, 8,2);	/* begin color blue */
	dp ("Exposing blue...");
	for (y=0; y<ymax; y++)
	{
		for (x=0; x<xmax; x++) buf[x]= (x*y) & 0xff;
		fr_data (fid, buf, xmax);
		if (y==200) 		/* test the skip source lines command */
		{
			fr_command (fid,10,200);
			y+=200;
		}
	}
	fr_command (fid, 9,0);	/* end color */
	dp ("done\n");

	fr_command (fid, 6,0);	/* end exposure */

	get_errors (fid);
			
	
}


/********************************************************************/

/* This little piece of code is intended to perform some communication
   with the FR-1 monitor through a file. It is not perfect, but
   for now, it does the job. */
   
void tell_monitor (unsigned char* message)
{
FILE* FP;
char fname [256];

	strcpy (fname, tmpname);
	strcat (fname, "/fr1.msg");
	FP = fopen (fname, "wt");
	if (FP!=NULL)
	{
		fprintf (FP, message);
		fclose (FP);
	}
}
	


/********************************************************************/

/* 	The following section takes care of the input file processing.
	It reads a ppm file (portable 24-bit raw format). The ppm format is:
	P6 (0x50, 0x36, 0x0a)		Magic number
	xres (space) yres (lf)		X and Y size
	maxval (lf)					maximum value for autoscaling
	# comment (lf)				one or more lines starting with #
	Binary data (3 bytes r-g-b) times X*Y
*/


unsigned char skiplf (FILE* fp)
{
unsigned char c;

	do	/* skip lfs, spaces, tabs etc */
	{
		fread (&c,1,1,fp);
	} 
	while ((!feof (fp)) && ((c==' ')||(c=='\n')||(c=='\t')||(c=='\r')));
	return c;
}


unsigned char skipnonlf (FILE* fp)
{
unsigned char c;

	do	/* skip to the next lf - not including spaces etc */
	{
		fread (&c,1,1,fp);
	} 
	while ((!feof (fp)) && (c!='\n') && (c!='\r'));
	return c;
}


int ppm_open ()
{
unsigned char c;
short magicnum;
int dummy;

	rows=cols=maxgray=0; magicnum=0;
	
	dp ("Reading header of file %s ", fname);
	fread (&magicnum, 2, 1, inp); /* rather have 2 bytes, 1 element fixed */
	if (feof (inp)) exit (0);		/* May have arrived at end of file */
	
	if (magicnum!=0x3650) /* Magic number 'P6'=0x50 0x36 */
	{
		dp ("Error: File with header %c%c (%x) is not a ppmraw file\n",
				magicnum & 0xff, magicnum >>8, magicnum);
		exit (1);
	}
	
	c = skiplf (inp);
	while (c=='#')			/* if comment, advance to the 1st char of the next line */
	{
		c=skipnonlf (inp); 
		c=skiplf (inp); 
	}
	if ((c>='0') && (c<='9'))		/* the start of a number */
	{
		rows= c & 0x0f;
		do
		{
			fread (&c, 1,1,inp);
			if ((c>='0') && (c<='9')) rows=rows*10+ (c & 0x0f);
		}
		while ((!feof (inp)) && (c>='0') && (c<='9'));
		c = skiplf (inp);
	}
	while (c=='#')			/* if comment, advance to the 1st char of the next line */
	{
		c=skipnonlf (inp);
		c=skiplf (inp);
	}
	if ((c>='0') && (c<='9'))		/* the start of a number */
	{
		cols= c & 0x0f;
		do
		{
			fread (&c, 1,1,inp);
			if ((c>='0') && (c<='9')) cols=cols*10+ (c & 0x0f);
		}
		while ((!feof (inp)) && (c>='0') && (c<='9'));
		c = skiplf (inp);
	}
	while (c=='#')			/* if comment, advance to the 1st char of the next line */
	{
		c=skipnonlf (inp);
		c=skiplf (inp);
	}
	if ((c>='0') && (c<='9'))		/* the start of a number */
	{
		maxgray= c & 0x0f;
		do
		{
			fread (&c, 1,1,inp);
			if ((c>='0') && (c<='9')) maxgray = maxgray*10+ (c & 0x0f);
		}
		while ((!feof (inp)) && (c>='0') && (c<='9'));
	}

	if (feof (inp)) 
	{
		dp ("Error: EOF within header\n");
		exit (1);
	}
	
	if ((rows==0) || (cols==0) || (maxgray==0))
	{
		dp ("Image header error: Read size %d x %d, max value %d\n",rows,cols,maxgray);
		exit (1);
	}
	flip = 0;
	if (rows < cols) 	/* portrait pixmap given */
	{
		flip = 1; dummy=rows; rows=cols; cols=dummy;	/* Swap x and y */
	}
		
	
	/* Finally, allocate memory */


	if (memhog)
	{
		PANE = (unsigned char*) calloc (rows*cols*3, sizeof (unsigned char));
		if (PANE==0)
		{
			memhog=0; autoframe=0;
			dp ("Memhog mode turned off - not enough memory\n");
		}
	}
	
	if ((!memhog) && (PANE==NULL))
	{
		PANE = (unsigned char*) calloc (rows*cols, sizeof (unsigned char));
		if (PANE==0)
		{
			dp ("Error (most fatal): Not enough memory, %d kB needed\n",rows*cols/1024);
			exit (1);
		}
	}
	
	hdrend = ftell (inp); 	/* Save the current position in the file */
	
	/* Finally, calculate the scaling factor */
	
	begx=0; begy=0; endx=rows; endy=cols; /* Have them defined even in non-autoframe mode */
	scalefac = fmax ((double)rows/(double)xmax, (double)cols/(double)ymax);
	padx = 0.5*(scalefac*xmax-rows); /* Positive pad values add to the image */
	pady = 0.5*(scalefac*ymax-cols); /* Negative values would crop */
	deltax=0; deltay=0;

	dp ("(%d x %d, max %d, pos %d). SUCCESS!\n", rows, cols, maxgray,hdrend);

	if (flip) dp ("Flipped ");
	dp ("Scalefac=%6.3f, padX=%6.3f, padY=%6.3f\n", scalefac,padx,pady);

	return 0;		/* We cancel the program anyway if there is an error in here */	
}



void readpane (char color)		/* Read the pane for the actual color */
{
int x,y,k;
int err;
unsigned char *rgb;		/* one row or col of RGB triplets */
long rc,rc2,ry;			/* Variables to avoid multiplication operations */
long p1,p2,p3,n;				/* same */
unsigned char s[255];

	err=fseek (inp, hdrend, SEEK_SET);	/* Reposition to beg of data area */
	if (err)
	{
		dp ("Oops - fseek causes trouble?? Bad... (filepos=%d)\n",hdrend);
		exit (1);
	}

	rc = rows*cols;
	rc2 = rows*cols*2;
	
	/* Now that we know the size, set aside some memory to load
	   a complete row or column of rgb triplets, which speeds up
	   reading the file (over reading triplet by triplet) */
	   
	if (flip) p1=cols; else p1=rows;
	rgb = (unsigned char*) calloc (3*p1, sizeof (unsigned char));

	sprintf (s, "Reading frame %d\n", slideno+1);
	tell_monitor (s);
	/* memset (PANE, 0, rows*cols);    /* Default is black */
	if (flip)						/* Flip: y is the faster index */
	{
		x=rows-1;
		while ( (!feof (inp)) && (x>=0) )
		{
			y = 0; k=0; ry=0; p1=x; p2=x+rc; p3=x+rc2;
			n=fread (rgb, 1,3*cols, inp);	/* Read one column of rgb data */
			
			if (n<3*cols) 
			{
				dp ("Row %d Warning: Too few items read: %d instead of %d\n",x,n,3*cols);
				dp ("Feof is %d, ferror is %d, errno is %d\n",feof(inp), ferror(inp),errno);
			}
			
			while ( y<cols)
			{
				if (memhog)
				{
					PANE[p1 + ry] = rgb[k++];
					PANE[p2 + ry] = rgb[k++];
					PANE[p3 + ry] = rgb[k++];
				}
				else
					PANE[x + ry] = rgb[y*3+color];
				y++;
				ry +=rows;
			}
			x--;
		}
	}
	else						/* No flip: x is the faster index */
	{
		y=0;
		while ( (!feof (inp)) && (y<cols) )
		{
			x = 0; ry=rows*y; k=0;
			p1=x+ry; p2=x+ry+rc; p3=x+ry+rc2;
			fread (rgb, 1, 3*rows, inp);

			while (x<rows)
			{
				if (memhog)
				{
					PANE[p1++] = rgb[k++];
					PANE[p2++] = rgb[k++];
					PANE[p3++] = rgb[k++];
				}
				else
					PANE[p1++] = rgb[x*3+color];
				x++;
			}
			y++;
		}
	}
	free (rgb);
}




/*	Interpolation: We have a matrix of rows x cols values, and
	the slide printer accepts xmax x ymax. This is converted using
	the bilinear interpolation.
	In order to maintain isotropy, we must have a common scale factor.
	Scale factor and offset (pad) are determined when reading the header
	or - if requested - in detect_white_border ()
*/


int interpol (int x, int y, int color)   /* Bi-linear interpolation */
{
int a,b,c,d,ix,iy;
double r,rx,ry;
long p;

	rx = scalefac*x-padx; ry = scalefac*y-pady;
	if ((rx<0) || (ry<0)) return 0;

	ix = (int) (rx); iy = (int) (ry);
	rx = rx-ix; ry = ry-iy;
	
	if ((ix>=endx-1) || (iy>=endy-1)) return 0;
	
	if (memhog) p=color*rows*cols; else p=0;
	p = p + ix + rows*iy;	/* index to 1st pixel of the square */

	a = PANE[p];
	b = PANE[p+1];
	c = PANE[p+rows];
	d = PANE[p+rows+1];
	r = a*(1-rx)*(1-ry)+b*rx*(1-ry)+c*(1-rx)*ry+d*rx*ry;
	if (r>255.49) r=255.49;

	return (int) (0.5+r);
}



/* For verification purposes: Write pane as pgmraw */

void write_pgm (int color)
{
FILE* pgmp;
unsigned char s[255], fn[255];
int x,y,x1,y1;
unsigned char buf[4096];

	switch (color)
	{
		case 0: 
		{ 
			strcpy (fn,tmpname); strcat (fn, "/red.pgm");
			strcpy (s,"P5\n#FR-1 red pane\n");
			break;
		}
		case 1: 
		{ 
			strcpy (fn,tmpname); strcat (fn, "/green.pgm");
			strcpy (s,"P5\n#FR-1 green pane\n");
			break;
		}
		case 2: 
		{ 
			strcpy (fn,tmpname); strcat (fn, "/blue.pgm");
			strcpy (s,"P5\n#FR-1 blue pane\n");
			break;
		}
	}
	
	pgmp = fopen (fn, "wb");
	if (pgmp==0) return;
	
	fwrite (s, strlen(s), 1, pgmp);	/* header & comment */
	sprintf (s, "%d %d\n%d\n", xmax, ymax, 255);
	fwrite (s, strlen(s), 1, pgmp); /* size & grayvalue */

	/* The following must be an exact mirror of the code found within
	   the exposure subroutine. Make sure it matches!
	   Note: The fr_data command is replaced by a fwrite */

	y1 = deltay/2;
	if (y1>0)
	{
		for (y=0; y<y1; y++)	/* Skip these lines */
		{
			for (x=0; x<xmax; x++) buf[x]= 0;
			fwrite (&buf, xmax,1, pgmp);
		}
	}
	else
		y1=0;

	for (y=y1; y<ymax; y++)
	{
		x1 = deltax/2;		/* padding to the left */
		if (x1>0)
		{
			for (x=0; x<x1; x++) buf[x]=0;
		}
		else
			x1=0;
	
		for (x=x1; x<xmax; x++) buf[x]= interpol (x-x1,y-y1,color);
		fwrite (&buf, xmax,1, pgmp);
	}
	fclose (pgmp);

}


/* write_thumbnail outputs a thumbnail of the current frame
   in pgm format (memhog mode only). It creates one or two
   files, depending on the setting of traceflg. The first file,
   always named "thumb.ppm" is used by the monitor program,
   while the second file, "thumbxxx.ppm" is created as a
   frame trace */

void write_thumbnail ()
{
int x,y;
char s[255];
char s1[255];
char s2[255];
unsigned char rgb[3];
int facx, facy;
FILE* th1;
FILE* th2;


	if (!memhog) return;
	
	sprintf (s1, "%s/thumb.ppm",tmpname);
	th1 = fopen (s1, "wb");
	if (th1==0) return;
	th2 = NULL; 		/* to silence a compiler warning */

	if (traceflg)
	{
		sprintf (s2, "%s/thumb%03d.ppm",tmpname,slideno+1);
		th2 = fopen (s2, "wb");
		if (th2==0) return;
	}


	sprintf (s,"P6\n#FR-1 thumbnail frame %d\n",slideno+1);
	fwrite (s, strlen(s), 1, th1);	/* header & comment */
	if (traceflg) fwrite (s, strlen(s), 1, th2);
	sprintf (s, "%d %d\n%d\n", thumb_x, thumb_y, 255);
	fwrite (s, strlen(s), 1, th1); /* size & grayvalue */
	if (traceflg) fwrite (s, strlen(s), 1, th2);

	facx = (xmax-deltax) / thumb_x;
	facy = (ymax-deltay) / thumb_y;

	for (y=0; y<thumb_y; y++)
		for (x=0; x<thumb_x; x++)
		{
			rgb[0] = interpol (x*facx, y*facy, 0);
			rgb[1] = interpol (x*facx, y*facy, 1);
			rgb[2] = interpol (x*facx, y*facy, 2);
			fwrite (&rgb, 1, sizeof (rgb), th1);
			if (traceflg) fwrite (&rgb, 1, sizeof (rgb), th2);
		}

	fclose (th1);
	if (traceflg) fclose (th2);
	chmod (s1, 0x1b6);				/* octal 666 */
	if (traceflg) chmod (s2, 0x1b6);

}

/*

In memhog mode (only in memhog mode), we can try to detect and eliminate
white borders that may have been generated by printing a 24x36 slide
on a A4 (or worse, letter) paper. Adjust scalefac and padx/pady.
*/

#define WHITE	0xff
#define BORDSPC 5

int iswhite (int x, int y)	/* only valid in memhog mode */
{
int p,q;

	p = x + rows*y; q = rows*cols;
	if ( (PANE[p] ==WHITE)				/* look in red pane */
		 && (PANE[p + q]==WHITE)		/* Look in green pane */
		 && (PANE[p + q + q]==WHITE))	/* look in blue pane */
		return 1;
	else
		return 0;
}



void detect_white_borders()
{
int x,y;

	begx=0; begy=0; endx=rows; endy=cols; deltax=0; deltay=0;
	if (memhog==0) return;

	dp ("Trying to find a white border to eliminate...\n");
	
	y = 0; x = -1;
	do
	{
		x++; if (x>=rows) { x=0; y++; }
	}
	while ((y<cols) && (iswhite (x,y)));
	
	/* Borders only make sense if they are far away from the center
	   of the image, so check. If (for example) BORDSPC==5 this means
	   that the border has to be less than 20% of the image width
	   otherwise it will be ignored. */
	   
	if ((BORDSPC*x < rows) && (BORDSPC*y < cols)) 
		{ begx=x; begy=y; }
	else
		dp ("Upper / left margin exceeds limits\n");
	
	/* Now do the same for a possible upper border. We do not assume
	   a centered image */
	   
	y = cols-1; x = rows;
	do
	{
		x--; if (x==0) { x=rows-1; y--; }
	}
	while ((y>=0) && (iswhite (x,y)));
	if ( (BORDSPC*(x-rows) < rows) && (BORDSPC*(y-cols)<cols) )
		{ endx=x; endy=y; }
	else
		dp ("Right / lower white border exceeds limits\n");
	
	dp ("Found white frame. Image coordinates are (%d,%d) - (%d,%d)",
		begx,begy,endx,endy);
	dp (" within (0,0) - (%d,%d)\n",rows,cols);

	/* Finally, calculate the scaling factor - overwrite the previous one */
	
	scalefac = fmax ((double)(endx-begx)/(double)xmax, 
					 (double)(endy-begy)/(double)ymax);
					 
	/* padx and pady (image domain) are negative. Image will be cropped */
					 
	padx = -begx; pady = -begy;
	
	/* For cropped images, deltax and deltay (slide domain) determine
	   the padding */
	   
	deltax = (int)(0.5+xmax-(endx-begx)/scalefac);
	deltay = (int)(0.5+ymax-(endy-begy)/scalefac);
	
	dp ("(%d x %d, max %d, pos %d). SUCCESS!\n", rows, cols, maxgray,hdrend);

	if (flip) dp ("Flipped ");
	dp ("Scalefac=%7.4f, padX=%6.3f, padY=%6.3f\n", scalefac,padx,pady);
	dp ("Offsetting by (%d,%d)\n", deltax, deltay);
	

}
		
		

/*====================================================================

Here, we finally arrived at the real expose part. It initializes the
slide printer, sets the parameters, and exposes the three panes.

======================================================================*/

void expose (int fid)
{
int x,y, x1,y1, color;
unsigned char buf[4096];
unsigned char ftype;

	/* start with a reset (initialize, command 0).
		errors should be cleared. If errors occur, we may safely
		exit the program because it won't work anyway. Errors
		of this type include camera back not attached, or one of the
		pins 4 and 7 pulled low
	*/

	fr_command (fid, 0,0);
	switch (request_sense(fid))
	{
		case 0: dp ("Request_sense (1) indicated no errors\n"); break;
		case 1: dp ("Request_sense (1) indicated warnings\n"); 
				get_errors(fid);
				break;
		case 2: dp ("Request_sense (1) indicated fatal errors\n");
				get_errors(fid);
				if (!forceoper) return;
				break;
	}


	
	/* Set up parameters */
	
	fr_command (fid, 3, resolution);
	
	/* Film type: 	0 = default Ektachrome 100
					1 = Ektachrome 100
					2 = Polachrome instant 35 mm
					3 = Polaroid 339 paper film
					4 = Polaroid 669
					5 = Polaroid 691
					6 = Kodacolor 100
					
		OR into this a 16 if fine color selected
		OR into this a 32 if low contrast selected
	*/
	
	ftype = filmtype;
	if (finecolor) ftype |= 16;
	if (lowcontrast) ftype |= 32;
	fr_command (fid, 4, ftype);
	
	if (mirror) fr_command (fid, 7, 1);

	/* Quick-check for errs, 1=warning, 2=fatal 
	   Abort operation if any fatals occured during set-up */

	switch (request_sense(fid))
	{
		case 0: dp ("Request_sense (2) indicated no errors\n"); break;
		case 1: dp ("Request_sense (2) indicated warnings\n"); 
				get_errors (fid);
				break;
		case 2: dp ("Request_sense (2) indicated fatal errors\n");
				get_errors (fid);
				if (!forceoper) return;
				break;
	}
	
	switch (get_ready_status (fid))
	{
		case 0: dp ("Get_ready_status: Camera not ready for unknown reason\n");
				/* if (!forceoper) return; */
				break;
		case 1:	dp ("Get_ready_status: Device is ready.\n");
				break;
		case 2: dp ("Get_ready_status: Camera not ready - out of film\n");
				if (!forceoper) return;
				break;
		case 3: dp ("Get_ready_status: Camera not ready - No camera back attached\n");
				if (!forceoper) return;
				break;
	}
				
	
	

	/* Do the actual exposure */
	
	fr_command (fid, 5,0);   /* start exposure */

	for (color=0; color<=2; color++)	/* Repeat for the 3 panes */
	{	
	
		if (testmode) write_pgm (color);
		fr_command (fid, 8,color);	/* begin color red */
		switch (color)
		{
			case 0: dp ("Exposing red..."); break;
			case 1: dp ("Exposing green..."); break;
			case 2: dp ("Exposing blue..."); break;
		}		
		switch (color)
		{
			case 0: tell_monitor ("Exposing red\n"); break;
			case 1: tell_monitor ("Exposing green\n"); break;
			case 2: tell_monitor ("Exposing blue\n"); break;
		}		


		/* Make sure that changes in this section are mirrored
		   in write_pgm() so that the control pgms match */

		y1 = deltay/2;
		if (y1>0)
		{
			for (y=0; y<y1; y++)	/* Skip these lines */
			{
				for (x=0; x<xmax; x++) buf[x]= 0;
				fr_data (fid, buf, xmax);
			}
		}
		else
			y1=0;

		for (y=y1; y<ymax; y++)
		{
			x1 = deltax/2;		/* padding to the left */
			if (x1>0)
			{
				for (x=0; x<x1; x++) buf[x]=0;
			}
			else
				x1=0;
		
			for (x=x1; x<xmax; x++) buf[x]= interpol (x-x1,y-y1,color);
			fr_data (fid, buf, xmax);
		}
		fr_command (fid, 9,0);	/* end color */
		dp ("done\n");

		/* in non-memhog mode, get the next color pane into the buffer */
		if ((!memhog) && (color<2)) readpane (color+1); 

	}
	fr_command (fid, 6,0);	/* end exposure */

	switch (get_errors (fid))
	{
		case 1: tell_monitor ("FR1 returned warnings\n"); break;
		case 2: tell_monitor ("FR1 returned ERRORS\n"); break;
		case 3: tell_monitor ("FR1 had serious trouble\n"); break;
		default: tell_monitor ("FR1 idle\n");	
	}
}



/********************************************************************/



void initparms ()
{
FILE* pf;
char s[255];
char s1[10];
int line;
int i,j;

	/* Set everything to default paramters */

	resolution = 0;
	xmax = 2048; ymax = 1365;
	finecolor=1;
	lowcontrast=0;
	forceoper=0;
	memhog=0;
	testmode=0;
	autoframe=0;
	mirror=0;
	traceflg=0;
	filmtype =0;
	thumb_y = 200; thumb_x = (3*thumb_y)/2;	/* default thumbnail size */
	
	/* Try to open & read the config file, /etc/montage.conf */
	
	pf = fopen ("/etc/montage.conf", "r");
	if (!pf) return;	/* no file - nothing to do */
	
	line=0;
	while (fgets (s, 255, pf) != NULL)
	{
		line++;
		if ((!isempty(s)) && (s[0]!='#'))	/* not a comment line */
		{
			if (!strncmp (s, "hires", 5)) 
				{ resolution=1; xmax = 4096; ymax = 2731; }
			else if (!strncmp(s, "lores", 5))
				{ resolution=0; xmax = 2048; ymax = 1365; }
			else if (!strncmp(s, "7bit", 4)) finecolor=0;
			else if (!strncmp(s, "8bit", 4)) finecolor=1;
			else if (!strncmp(s, "debug", 5)) { debugmsg=1; debugdump=1; }
			else if (!strncmp(s, "quiet", 5)) { debugmsg=0; debugdump=0; }	
			else if (!strncmp(s, "verbose", 7)) { debugmsg=1; debugdump=1; }	
			else if (!strncmp(s, "memhog", 6)) memhog=1;
			else if (!strncmp(s, "test", 4)) testmode=1;
			else if (!strncmp(s, "force", 5)) forceoper=1;
			else if (!strncmp(s, "auto", 4)) autoframe=1;
			else if (!strncmp(s, "mirror",6)) mirror=1;
			else if (!strncmp(s, "lowcontrast",11)) lowcontrast=1;
			else if (!strncmp(s, "trace", 5)) traceflg=1;
			else if (!strncmp(s, "device", 6))
			{
				i=6;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (devname, &s[i], j-i);
				
			}
			else if (!strncmp(s, "input", 5))
			{
				i=5;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (fname, &s[i], j-i);
			}
			else if (!strncmp(s, "tempdir", 7))
			{
				i=7;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (tmpname, &s[i], j-i);
			}
			else if (!strncmp(s, "filmtype", 8))
			{
				i=7;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], j-i);
				filmtype = atoi (s1);
				if (filmtype>6)
				{
					dp ("Warning: Illegal film type %d in /etc/montage.conf, line %d\n", filmtype,line);
					filmtype=0;
				}
			}
			else
				dp ("Warning: Illegal keyword in /etc/montage.conf, line %d\n",line);

		}
	}				

	if ((!memhog) && autoframe)
	{
		dp ("Warning: Automatic frame detection turned off - not in memhog mode\n");
		autoframe=0;
	}
	
}

void prhelp ()
{
	fprintf (stderr, "FR-1: Printer filter for Montage FR-1 slide printer\n");
	fprintf (stderr, "%s (c) Mark A. Haidekker\n",version);
	fprintf (stderr, "valid options are:\n");
	fprintf (stderr, "-? -help          print this help text\n");
	fprintf (stderr, "-8 -8bit          8-bit color (slower, default)\n");
	fprintf (stderr, "-7 -7bit          7-bit color (faster)\n");
	fprintf (stderr, "-a -auto          Autodetect & eliminate white frame (Memhog mode only)\n");
	fprintf (stderr, "-d -device        specify device name (default /dev/sgb)\n");
	fprintf (stderr, "-f -force         force continuation despite fatal errors\n");
	fprintf (stderr, "   -filmtype      Set a filmtype (range 0..6, default: 0)\n");
	fprintf (stderr, "-h -hires         high 4096x2731 resolution\n");
	fprintf (stderr, "-i -input         specify input file (default stdin)\n");
	fprintf (stderr, "-l -lores         low  2048x1362 resolution (default)\n");
	fprintf (stderr, "-m -memhog        Buffer all three panes (40 MB RAM!)\n");
	fprintf (stderr, "   -mirror        Mirror the slide (FR-1 function)\n");
	fprintf (stderr, "-q -quiet         Don't print any messages\n");
	fprintf (stderr, "-s -soft          Soft (low) contrast mode\n");
	fprintf (stderr, "-t -test          Write the color panes as pgm to /var/tmp\n");
	fprintf (stderr, "   -tempdir       Specify temp directory (default: /var/tmp\n");
	fprintf (stderr, "   -trace         Create a thumbnail trace of all exposed frames\n");
	fprintf (stderr, "-v -verbose       Be verbose - cancel quiet mode\n");
}


void print_parameter_info ()
{

	dp ("you selected the following parameters:\n");
	dp ("Device: %s\n", devname);
	dp ("Input file: %s\n", fname);
	dp ("Temporary files will be created under %s\n", tmpname);

	dp ("Resolution: %d by %d\n", xmax,ymax);
	dp ("Film type is %d\n", filmtype);

	if (finecolor) dp ("Fine (8-bit) color depth\n"); else dp ("Low (7-bit) color depth\n");
	if (lowcontrast) dp ("Low contrast setting\n"); else dp ("Normal contrast setting\n");
	if (mirror) dp ("image will be mirrored\n"); else dp ("No mirroring of the slide\n");
	if (forceoper) dp ("Operation through fatal errors enforced\n");
	if (memhog) dp ("Memory hog option is ON\n");
	if (autoframe) dp ("White frame will be detected and eliminated\n");
	if (testmode) dp ("Test mode: Creates %s/red.pgm etc for review\n",tmpname);
	if (traceflg) dp ("Creating thumbnail trace\n");
	dp ("=====================================================\n");

}


/********************************************************************/


#define nextargi (--argc,atoi(*++argv))
#define nextargf (--argc,atof(*++argv))
#define nextargs (--argc,*++argv)



/* Hint: Make sure that /dev/sgx (whatever the generic device) is chmod a+rw
   - otherwise, only root can use this program */

int main(int argc,char **argv)
{
int fd;
char model[256], version[256];

	strcpy (version, "v 0.9 alpha");
	strcpy (devname, "/dev/sgb");
	strcpy (fname, "stdin");
	strcpy (tmpname, "/var/tmp");
	initparms ();
	
    while (--argc && **++argv=='-') 
	{

		if (!strcmp(argv[0],"-h") || !strcmp(argv[0],"-hires"))
			{ resolution=1; xmax = 4096; ymax = 2731; }

		else if (!strcmp(argv[0],"-l") || !strcmp(argv[0],"-lores"))
			{ resolution=0; xmax = 2048; ymax = 1365; }

		else if (!strcmp(argv[0],"-8") || !strcmp(argv[0],"-8bit"))
			finecolor=1;

		else if (!strcmp(argv[0],"-7") || !strcmp(argv[0],"-7bit"))
			finecolor=0;
			
		else if (!strcmp(argv[0],"-f") || !strcmp(argv[0],"-force"))
			forceoper=1;

		else if (!strcmp(argv[0],"-m") || !strcmp(argv[0],"-memhog"))
			memhog=1;

		else if (!strcmp(argv[0],"-t") || !strcmp(argv[0],"-test"))
			testmode=1;

		else if (!strcmp(argv[0],"-trace"))
			traceflg=1;

		else if (!strcmp(argv[0],"-a") || !strcmp(argv[0],"-auto"))
			autoframe=1;

		else if (!strcmp(argv[0],"-s") || !strcmp(argv[0],"-soft"))
			lowcontrast=1;

		else if (!strcmp(argv[0],"-d") || !strcmp(argv[0],"-dev"))
			strcpy (devname,nextargs);

		else if (!strcmp(argv[0],"-q") || !strcmp(argv[0],"-quiet"))
			{ debugmsg=0; debugdump=0; }

		else if (!strcmp(argv[0],"-v") || !strcmp(argv[0],"-verbose"))
			{ debugmsg=1; debugdump=1; }

		else if (!strcmp(argv[0],"-i") || !strcmp(argv[0],"-input"))
			strcpy (fname,nextargs);

		else if (!strcmp(argv[0],"-tempdir"))
			strcpy (tmpname,nextargs);

		else if (!strcmp(argv[0],"-mirror"))
			mirror=1;

		else if (!strcmp(argv[0],"-filmtype"))
			filmtype=nextargi;

        else 
		{
        	if (strcmp(argv[0],"-help") && strcmp (argv[0],"-?"))
            	fprintf(stderr,"unknown option '%s'; \n",argv[0]);
			prhelp ();

			exit (1);
		}
	}

	print_parameter_info ();
	dp ("FR-1: Printer filter for Montage FR-1 slide printer\n");
	dp ("%s (c) Mark A. Haidekker\n",version);

	PANE=NULL;		/* Unallocated memory should be null pointer */
	
	
	/* If gs gets passed multiple pages, it just adds one after another
	   of the converted images to the same file. In this case,
	   we expect that after the last byte of the n-th picture
	   there is the next "P6" of the n+1-th picture.
	   Therefore, we initialize everything once (i.e. open
	   input file, then open sg device) and loop until we run
	   out of bytes 
	*/
	
	inp = fopen (fname, "rb");
	if (inp==NULL) 
	{
		dp ("Error opening input file %s\n",fname);
		tell_monitor ("File open failed");
		exit (1);	/* Crash. No way. */
	}

    fd = open (devname, O_RDWR);   /* open is in fcntl.h */
	if (fd<0)
	{
		if (debugmsg) fprintf (stderr, "Problem %d opening %s\n",fd,devname);
		if (debugmsg) fprintf (stderr, "Device may be write-protected\n");
		if (debugmsg) perror ("System error code ");
		return -1;
	}
	
	dp ("open successful\n");
    inquire_model (fd, model, version);
	dp ("Device found: %s version %s\n", model, version);
	
	slideno=0;
	while (!feof (inp))
	{
		ppm_open ();	/* Reads the header and allocates memory */
		dp ("Frame %d: Filling input buffer...",slideno+1);
		readpane (0);	/* Reads first pane BEFORE exposure */
		dp ("Done.\n");

		if (memhog && autoframe) detect_white_borders();

		write_thumbnail ();

		wait_for_ready (fd);		/* wait until CHECK_COND vanishes
									   and the device signals readyness.
									   wait_for_ready should be used sparingly */

		expose (fd);
		if (PANE!=NULL) free (PANE);	/* Discard memory */
		slideno++;
	}


    close(fd);
	fclose (inp);
	dp ("Total number of frames exposed: %d\n", slideno);

	return 0;
}

